@@ -54,6 +54,8 @@ put_delimiter(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type)
   case GRN_CONTENT_MSGPACK :
     // do nothing
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
@@ -88,6 +90,8 @@ grn_output_array_open(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_typ
     msgpack_pack_array(&ctx->impl->msgpacker, nelements);
 #endif
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
@@ -119,6 +123,8 @@ grn_output_array_close(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_ty
   case GRN_CONTENT_MSGPACK :
     // do nothing
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
@@ -155,6 +161,8 @@ grn_output_map_open(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type,
     msgpack_pack_map(&ctx->impl->msgpacker, nelements);
 #endif
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
@@ -186,6 +194,8 @@ grn_output_map_close(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type
   case GRN_CONTENT_MSGPACK :
     // do nothing
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
@@ -214,6 +224,9 @@ grn_output_int32(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type, in
     msgpack_pack_int32(&ctx->impl->msgpacker, value);
 #endif
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    grn_text_itoa(ctx, outbuf, value);
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
@@ -241,6 +254,9 @@ grn_output_int64(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type, in
     msgpack_pack_int64(&ctx->impl->msgpacker, value);
 #endif
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    grn_text_lltoa(ctx, outbuf, value);
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
@@ -268,6 +284,9 @@ grn_output_uint64(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type, i
     msgpack_pack_uint64(&ctx->impl->msgpacker, value);
 #endif
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    grn_text_ulltoa(ctx, outbuf, value);
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
@@ -295,6 +314,9 @@ grn_output_float(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type, do
     msgpack_pack_double(&ctx->impl->msgpacker, value);
 #endif
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    grn_text_ftoa(ctx, outbuf, value);
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
@@ -324,6 +346,9 @@ grn_output_str(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type,
     msgpack_pack_raw_body(&ctx->impl->msgpacker, value, value_len);
 #endif
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    GRN_TEXT_PUT(ctx, outbuf, value, value_len);
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
@@ -362,6 +387,9 @@ grn_output_bool(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type, grn
     }
 #endif
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    GRN_TEXT_PUTS(ctx, outbuf, value ? "true" : "false");
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
@@ -386,6 +414,8 @@ grn_output_null(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type)
     msgpack_pack_nil(&ctx->impl->msgpacker);
 #endif
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
@@ -426,6 +456,9 @@ grn_output_time(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type, int
     msgpack_pack_double(&ctx->impl->msgpacker, dv);
 #endif
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    grn_text_ftoa(ctx, outbuf, dv);
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
@@ -487,6 +520,17 @@ grn_output_geo_point(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type
     }
 #endif
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    if (value) {
+      GRN_TEXT_PUTC(ctx, outbuf, '"');
+      grn_text_itoa(ctx, outbuf, value->latitude);
+      GRN_TEXT_PUTC(ctx, outbuf, 'x');
+      grn_text_itoa(ctx, outbuf, value->longitude);
+      GRN_TEXT_PUTC(ctx, outbuf, '"');
+    } else {
+      GRN_TEXT_PUTS(ctx, outbuf, "\"\"");
+    }
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
@@ -519,10 +563,10 @@ grn_text_atoj(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type,
         buf.header.domain = DB_OBJ(a->obj)->range;
         break;
       case GRN_ACCESSOR_GET_SCORE :
-        grn_obj_get_value(ctx, a->obj, id, &buf);
         {
           grn_rset_recinfo *ri = (grn_rset_recinfo *)grn_obj_get_value_(ctx, a->obj, id, &vs);
-          GRN_INT32_PUT(ctx, &buf, ri->score);
+          int32_t int32_score = ri->score;
+          GRN_INT32_PUT(ctx, &buf, int32_score);
         }
         buf.header.domain = GRN_DB_INT32;
         break;
@@ -1141,6 +1185,20 @@ count_used_n_codes(grn_ctx *ctx, grn_expr_code *start, grn_expr_code *target)
   return n_codes;
 }
 
+static grn_bool
+is_score_accessor(grn_ctx *ctx, grn_obj *obj)
+{
+  grn_accessor *a;
+
+  if (obj->header.type != GRN_ACCESSOR) {
+    return GRN_FALSE;
+  }
+
+  for (a = (grn_accessor *)obj; a->next; a = a->next) {
+  }
+  return a->action == GRN_ACCESSOR_GET_SCORE;
+}
+
 static inline void
 grn_output_table_column(grn_ctx *ctx, grn_obj *outbuf,
                         grn_content_type output_type,
@@ -1148,13 +1206,16 @@ grn_output_table_column(grn_ctx *ctx, grn_obj *outbuf,
 {
   grn_output_array_open(ctx, outbuf, output_type, "COLUMN", 2);
   if (column) {
-    grn_id range_id;
+    grn_id range_id = GRN_ID_NIL;
     GRN_BULK_REWIND(buf);
     grn_column_name_(ctx, column, buf);
     grn_output_obj(ctx, outbuf, output_type, buf, NULL);
     if (column->header.type == GRN_COLUMN_INDEX) {
       range_id = GRN_DB_UINT32;
-    } else {
+    } else if (is_score_accessor(ctx, column)) {
+      range_id = GRN_DB_INT32;
+    }
+    if (range_id == GRN_ID_NIL) {
       range_id = grn_obj_get_range(ctx, column);
     }
     if (range_id == GRN_ID_NIL) {
@@ -1294,7 +1355,7 @@ grn_output_table_columns_by_columns(grn_ctx *ctx, grn_obj *outbuf,
   grn_output_array_close(ctx, outbuf, output_type);
 }
 
-static inline void
+void
 grn_output_table_columns(grn_ctx *ctx, grn_obj *outbuf,
                          grn_content_type output_type,
                          grn_obj *table, grn_obj_format *format)
@@ -1407,7 +1468,7 @@ grn_output_table_records_by_columns(grn_ctx *ctx, grn_obj *outbuf,
   }
 }
 
-static inline void
+void
 grn_output_table_records(grn_ctx *ctx, grn_obj *outbuf,
                          grn_content_type output_type,
                          grn_obj *table, grn_obj_format *format)
@@ -2006,7 +2067,57 @@ grn_output_envelope(grn_ctx *ctx,
     }
 #endif
     break;
+  case GRN_CONTENT_GROONGA_COMMAND_LIST :
+    break;
   case GRN_CONTENT_NONE:
     break;
   }
 }
+
+static inline grn_bool
+is_output_columns_format_v1(grn_ctx *ctx,
+                            const char *output_columns,
+                            unsigned int output_columns_len)
+{
+  unsigned int i;
+
+  /* TODO: REMOVE ME. If new output_columns handler is marked as stable,
+     this check is removed. We need more error checks. */
+  if (grn_ctx_get_command_version(ctx) == GRN_COMMAND_VERSION_1) {
+    return GRN_TRUE;
+  }
+
+  for (i = 0; i < output_columns_len; i++) {
+    switch (output_columns[i]) {
+    case ',' :
+    case '(' :
+    case '[' :
+      return GRN_FALSE;
+    default :
+      break;
+    }
+  }
+
+  return GRN_TRUE;
+}
+
+grn_rc
+grn_output_format_set_columns(grn_ctx *ctx, grn_obj_format *format,
+                              grn_obj *table,
+                              const char *columns, int columns_len)
+{
+  grn_rc rc;
+
+  if (is_output_columns_format_v1(ctx, columns, columns_len)) {
+    rc = grn_obj_columns(ctx, table, columns, columns_len, &(format->columns));
+  } else {
+    grn_obj *variable;
+    GRN_EXPR_CREATE_FOR_QUERY(ctx, table, format->expression, variable);
+    rc = grn_expr_parse(ctx, format->expression,
+                        columns, columns_len, NULL,
+                        GRN_OP_MATCH, GRN_OP_AND,
+                        GRN_EXPR_SYNTAX_OUTPUT_COLUMNS);
+  }
+
+  return rc;
+}
